/*
 * Copyright (c) 2005, 2009 Sun Microsystems, Inc. All Rights Reserved.
 * Use is subject to license terms.
 */
/*
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.01 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.php.net/license/3_01.txt                                  |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
   | Author: Basant Kukreja <basantk@php.net>                             |
   +----------------------------------------------------------------------+
 */

#include "php.h"

#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>

#ifndef HAVE_PHP_SESSION
#error Session support is required for memsession extension
#endif

#ifndef ZTS
#error memsession extension only works for threaded php.
#endif

#if !defined(PHP_WIN32) && !defined(PTHREADS)
#error pthread support is required for memsession extension
#endif

#include "php_session.h"
#include "php_memsession.h"
#include "main/SAPI.h"
#include "php_ini.h"


static zend_rwlock_t* _session_lock = NULL;

static const int _numLocks = 16;
static HashTable** _session_hash = NULL;
static const char* _sessfile_header = "In-memory session dump\n";

#if 0
#define ps_mems_debug(val) fprintf val
#else
#define ps_mems_debug(val)
#endif

ps_module ps_mod_memsession = {
	PS_MOD(memsession)
};

ZEND_DECLARE_MODULE_GLOBALS(memsession)

static int choose_lock_index(const char* key)
{
	int index = 0;
	switch (*key) {
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			index = *key - '0';
			break;
		case 'a':
		case 'b':
		case 'c':
		case 'd':
		case 'e':
		case 'f':
			index = *key - 'a' + 10;
			break;
		default:
			index = 15;
			break;
	}
	return index;
}

static void init_session_lock()
{
	int index = 0;
	_session_lock = (zend_rwlock_t*) malloc(_numLocks * sizeof(zend_rwlock_t));
	for (index = 0; index < _numLocks; ++index) {
		zend_init_rwlock(&_session_lock[index]);
	}
}

static void acquire_read_lock(int index)
{
    zend_acquire_read_lock(&_session_lock[index]);
}

static void acquire_write_lock(int index)
{
    zend_acquire_write_lock(&_session_lock[index]);
}

static void release_read_lock(int index)
{
    zend_release_read_lock(&_session_lock[index]);
}

static void release_write_lock(int index)
{
    zend_release_write_lock(&_session_lock[index]);
}

static int free_entry(memsession_data* sess_data)
{
	int ret = FAILURE;
	if (zend_atomic_dec_32_nv(&sess_data->refcount) == 0) {
		ps_mems_debug((stderr, "Free entry data = %x:%x refcount = %d\n",
					  sess_data, sess_data->val,
					  sess_data->refcount));
		if (sess_data->val) {
			free(sess_data->val);
			sess_data->val = NULL;
		}
		free(sess_data);
		ret = SUCCESS;
	}
	return ret;
}

PS_OPEN_FUNC(memsession)
{
	ps_memsession *data;

	data = emalloc(sizeof(*data));
	memset(data, 0, sizeof(*data));
	data->ctx = 1; /* Non null value */

	PS_SET_MOD_DATA(data);

	return SUCCESS;
}

PS_CLOSE_FUNC(memsession)
{
	ps_memsession *data = PS_GET_MOD_DATA();
	zend_bool ctx = data->ctx;

	if (ctx) {
		/* nothing to do to close session */
	}

	efree(data);
	*mod_data = NULL;

	return SUCCESS;
}

PS_READ_FUNC(memsession)
{
	ps_memsession *data = PS_GET_MOD_DATA();
	zend_bool ctx = data->ctx;
	int ret = SUCCESS;

	if (ctx) {
		memsession_data** psess_data = NULL;
		int index = choose_lock_index(key);
		int keylen = strlen(key);
		ulong hval = zend_inline_hash_func((char*) key, keylen);
		acquire_read_lock(index);
		zend_hash_quick_find(_session_hash[index], (char*) key, keylen,
							 hval, (void**) &psess_data);

		if (psess_data && *psess_data) {
			memsession_data* sess_data = *psess_data;
			zend_atomic_inc_32(&sess_data->refcount);
		}
		release_read_lock(index);
		if (psess_data && *psess_data) {
			memsession_data* sess_data = *psess_data;
			*vallen = sess_data->vallen;
			if (*vallen) {
				*val = emalloc(*vallen);
				memcpy(*val, sess_data->val, *vallen);
			} else {
				*val = STR_EMPTY_ALLOC();
			}
			/* Done with reading. */
			ps_mems_debug((stderr, "Reader doing free\n"));
			free_entry(sess_data);
		}
		else {
			ret = FAILURE;
		}
	}

	return ret;
}

PS_WRITE_FUNC(memsession)
{
	ps_memsession *data = PS_GET_MOD_DATA();
	zend_bool ctx = data->ctx;
	int ret = SUCCESS;

	if (data->ctx) {
		memsession_data* sess_data = NULL;
		memsession_data** psess_data = NULL;
		memsession_data* newdata;
		time_t now;
		int completed = 0;
		int keylen = strlen(key);
		ulong hval;
		int index = choose_lock_index(key);
		HashTable* ht = _session_hash[index];
#if PHP_API_VERSION < 20041225
#if HAVE_APACHE
		now = ((request_rec *)SG(server_context))->request_time;
#else
		now = time(0);
#endif
#else
		now = sapi_get_request_time(TSRMLS_C);
#endif

		newdata = malloc(sizeof(memsession_data));
		newdata->refcount = 1;
		newdata->vallen = vallen;
		newdata->val = malloc(newdata->vallen);
		memcpy(newdata->val, val, vallen);
		newdata->last_update_time = now;
		sess_data = NULL;
		hval = zend_inline_hash_func((char*) key, keylen);
		ps_mems_debug((stderr, "Going to new data %p:%p key = %s hval = %ld\n",
					  newdata, newdata->val, key, hval));

		acquire_write_lock(index);
		if (zend_hash_quick_find(ht, (char*) key, keylen,
								 hval, (void**) &psess_data) == SUCCESS) {
			if (psess_data && *psess_data) {
				sess_data = *psess_data;
			}
		}
		ret = zend_hash_quick_update(ht, (char*) key, keylen, hval, &newdata,
									 sizeof(void*), 0);
		release_write_lock(index);
		if (ret == SUCCESS) {
			if (sess_data) {
				ps_mems_debug((stderr, "writer doing free\n"));
				free_entry(sess_data);
			}
		} else {
			ps_mems_debug((stderr, "writer doing free newdata\n"));
			free_entry(newdata);
		}
		ps_mems_debug((stderr,
					   "session write : ret = %d psess_data = %x data=%x refcount= %d\n",
					   ret, sess_data, psess_data, (sess_data ? sess_data->refcount : -1)));
	}

	return ret;
}

static int delete_from_hash(HashTable* ht, char* key, int keylen,
					        memsession_data** out_sess_data)
{
	int ret = SUCCESS;
	memsession_data** psess_data = NULL;
	zend_hash_find(ht, (char*) key, strlen(key),
				   (void**) &psess_data);

	if (psess_data && *psess_data) {
		zend_hash_del(ht, (char*) key, strlen(key));
		*out_sess_data = *psess_data;
	}
	else {
		ret = FAILURE;
	}
	return ret;
}

PS_DESTROY_FUNC(memsession)
{
	ps_memsession *data = PS_GET_MOD_DATA();
	zend_bool ctx = data->ctx;
	int ret = SUCCESS;
	ps_mems_debug((stderr, "session destroyer \n"));
	if (data->ctx) {
		memsession_data* sess_data =  NULL;
		int index = choose_lock_index(key);
		acquire_write_lock(index);
		ret = delete_from_hash(_session_hash[index], (char*) key, strlen(key),
							   &sess_data);
		release_write_lock(index);
		if (sess_data) {
			ps_mems_debug((stderr, "destroyer doing free newdata\n"));
			if (free_entry(sess_data) != SUCCESS) {
				ps_mems_debug((stderr, "Destroyer unable to delete"));
			}
		}
	}

	return SUCCESS;
}

/* Structure used only by GC function PS_GC_FUNC */
typedef struct _memsession_key
{
	char* key;
	uint keylen;
	memsession_data* value;
} memsession_key;

PS_GC_FUNC(memsession)
{
	HashPosition pos;
	int cursize = 1024;
	time_t now;
	int index = 0;
	memsession_key* indexptr = NULL;
	time(&now);


	for (index = 0; index < _numLocks; ++index) {
		int countindex = 0;
		int keyindex = 0;
		acquire_write_lock(index);
		zend_hash_internal_pointer_reset_ex(_session_hash[index], &pos);
		while(zend_hash_has_more_elements_ex(_session_hash[index], &pos)
			   	== SUCCESS) {
			memsession_data** psess_data = NULL;
			zend_hash_get_current_data_ex(_session_hash[index],
										  (void **) &psess_data, &pos);
			if (psess_data && *psess_data) {
				memsession_data* sess_data = *psess_data;
				if (now - sess_data->last_update_time > maxlifetime) {
					ulong delindex;
					char* key = NULL;
					uint keylen = 0;
					int ret = zend_hash_get_current_key_ex(_session_hash[index],
														   &key,
														   &keylen,
														   &delindex, 0,
														   &pos);
					if (ret == HASH_KEY_IS_STRING) {
						if (indexptr == NULL) {
							indexptr = (memsession_key*) emalloc(
												cursize * sizeof(memsession_key));
						}
						if (countindex >= cursize) {
							cursize *= 2;
							indexptr = (memsession_key*) erealloc(indexptr,
												cursize * sizeof(memsession_key));
						}
						indexptr[countindex].key = key;
						indexptr[countindex].value = NULL;
						indexptr[countindex++].keylen = keylen;
					}
				}
			}
			zend_hash_move_forward_ex(_session_hash[index], &pos);
		}
		for (keyindex = 0; keyindex < countindex; ++keyindex) {
			delete_from_hash(_session_hash[index], indexptr[keyindex].key,
							 indexptr[keyindex].keylen,
							 &indexptr[keyindex].value);
		}
		release_write_lock(index);
		for (keyindex = 0; keyindex < countindex; ++keyindex) {
			memsession_data* sess_data = indexptr[keyindex].value;
			ps_mems_debug((stderr, "gc doing free\n"));
			if (sess_data) {
				if (free_entry(sess_data) != SUCCESS) {
					ps_mems_debug((stderr, "GC unable to delete"));
				}
			}
		}
		if (countindex > 0) {
			ps_mems_debug((stderr,
						  "memsession gc called : cleaned %d session entries\n",
						   countindex));
		}
	}
	if (indexptr != NULL) {
		efree(indexptr);
	}

	return SUCCESS;
}

static int restore_session_data(const char* save_path)
{
	char buf[80];
	int bytes_read;
	FILE* fp = fopen(save_path, "r");
	if (fp == NULL) {
		return SUCCESS;
	}
	if ((bytes_read = fread(buf, 1, strlen(_sessfile_header), fp))
						!= strlen(_sessfile_header)) {
		goto process_error;
	}
	if (memcmp(buf, _sessfile_header, strlen(_sessfile_header)) != 0)
		goto process_error;

	while (1) {
		memsession_data sess_data;
		memsession_data* newdata;
		memsession_data** psess_data = NULL;
		int keylen = 0;
		char* key = buf;
		ulong hval;
		int index = 0;
		HashTable* ht = NULL;
		memset(&sess_data, 0, sizeof(sess_data));
		sess_data.refcount = 1;
		/* Read key length */
		if (fread(&keylen, sizeof(int), 1, fp) < 1) {
			break;
		}
		if (keylen >= sizeof(buf)) {
			goto process_error;
		}
		/* Read key */
		if (fread(key, keylen, 1, fp) < 1)
			goto process_error;
		/* Read time */
		if (fread(&sess_data.last_update_time, sizeof(sess_data.last_update_time),
			      1, fp) < 1)
			goto process_error;
		/* Read session value length */
		if (fread(&sess_data.vallen, sizeof(sess_data.vallen),
			      1, fp) < 0)
			goto process_error;
		newdata = malloc(sizeof(memsession_data));
		memcpy(newdata, &sess_data, sizeof(memsession_data));
		newdata->val = malloc(newdata->vallen);
		if ((bytes_read = fread(newdata->val, 1, newdata->vallen,
			      			    fp)) < newdata->vallen) {
			free(newdata->val);
			free(newdata);
			goto process_error;
		}

		hval = zend_inline_hash_func((char*) key, keylen);
		index = choose_lock_index(key);
		ht = _session_hash[index];
		/* We don't need locks as init is always thread safe */
		if (zend_hash_quick_find(ht, (char*) key, keylen,
								 hval, (void**) &psess_data) != SUCCESS) {
			zend_hash_quick_update(ht, (char*) key, keylen, hval, &newdata,
										 sizeof(void*), 0);
		}
		else {
			/* Entry already exists. It should not happen */
			free(newdata->val);
			free(newdata);
			goto process_error;
		}
	}

	return SUCCESS;
process_error:
	fprintf(stderr, "Invalid session file\n");
	fclose(fp);
	return FAILURE;
}

static int save_session_data(const char* save_path)
{
	HashPosition pos;
	int index = 0;
	FILE* fp = fopen(save_path, "w");
	if (fp == NULL) {
		fprintf(stderr, "Unable to save session data (fopen failed for %s)\n",
				save_path);
		return FAILURE;
	}
	fprintf(fp, _sessfile_header);

	for (index = 0; index < _numLocks; ++index) {
		int countindex = 0;
		int keyindex = 0;
		acquire_write_lock(index);
		zend_hash_internal_pointer_reset_ex(_session_hash[index], &pos);
		while(zend_hash_has_more_elements_ex(_session_hash[index], &pos)
			   	== SUCCESS) {
			memsession_data** psess_data = NULL;
			zend_hash_get_current_data_ex(_session_hash[index],
										  (void **) &psess_data, &pos);
			if (psess_data && *psess_data) {
                ulong delindex;
                char* key = NULL;
                uint keylen = 0;
                int ret = zend_hash_get_current_key_ex(_session_hash[index],
                                                       &key,
                                                       &keylen,
                                                       &delindex, 0,
                                                       &pos);
                if (ret == HASH_KEY_IS_STRING) {
					memsession_data* sess_data = *psess_data;
					fwrite(&keylen, sizeof(keylen), 1, fp);
					fwrite(key, keylen, 1, fp);
					fwrite(&sess_data->last_update_time,
						  sizeof(sess_data->last_update_time), 1, fp);
					fwrite(&sess_data->vallen,
						   sizeof(sess_data->vallen), 1, fp);
					fwrite(sess_data->val, sess_data->vallen, 1, fp);
                }
			}
			zend_hash_move_forward_ex(_session_hash[index], &pos);
		}
		release_write_lock(index);
	}
	fclose(fp);

	return SUCCESS;
}

static void php_memsession_init_globals(zend_memsession_globals *memsession_globals_p TSRMLS_DC)
{
	MEMSESSION_G(save_path) = NULL;
}

PHP_INI_BEGIN()
	STD_PHP_INI_ENTRY("memsession.save_path",          "",          PHP_INI_ALL, OnUpdateString, save_path,          memsession_globals,    memsession_globals)
PHP_INI_END()

PHP_MINIT_FUNCTION(memsession)
{
	char* save_path;
	zend_atomic_init();
	ps_mems_debug((stderr, "memsession init function\n"));
	if (_session_hash == NULL) {
		int index = 0;
		_session_hash = malloc(_numLocks * sizeof(HashTable*));
		memset(_session_hash, 0, _numLocks * sizeof(HashTable*));
		for (index = 0; index < _numLocks; ++index) {
			_session_hash[index] = (HashTable*) malloc(sizeof(HashTable));
			zend_hash_init(_session_hash[index], 128, NULL, NULL, 1);
		}
		init_session_lock();
	}
	ts_allocate_id(&memsession_globals_id, sizeof(zend_memsession_globals),
				   (ts_allocate_ctor) php_memsession_init_globals, NULL);
	REGISTER_INI_ENTRIES();
	save_path = MEMSESSION_G(save_path);
	if (save_path && *save_path) {
		restore_session_data(save_path);
	}
	php_session_register_module(&ps_mod_memsession);
	return SUCCESS;
}

PHP_MSHUTDOWN_FUNCTION(memsession)
{
	char* save_path;
	save_path = MEMSESSION_G(save_path);
	if (save_path && *save_path) {
		save_session_data(save_path);
	}
	return SUCCESS;
}

PHP_MINIT_FUNCTION(ps_memsession);
PHP_MSHUTDOWN_FUNCTION(ps_memsession);

static const zend_module_dep memsession_deps[] = {
	ZEND_MOD_REQUIRED("session")
	{NULL, NULL, NULL}
};

zend_function_entry memsession_functions[] = {
	{NULL, NULL, NULL}
};

zend_module_entry memsession_module_entry = {
	STANDARD_MODULE_HEADER_EX,
	NULL,
	memsession_deps,
	"memsession",
	memsession_functions,
	PHP_MINIT(memsession),
	PHP_MSHUTDOWN(memsession),
	NULL,
	NULL,
	NULL,
	NO_VERSION_YET,
	STANDARD_MODULE_PROPERTIES
};

ZEND_GET_MODULE(memsession)
