/*
 * Copyright (c) 2005, 2009 Sun Microsystems, Inc. All Rights Reserved.
 * Use is subject to license terms.
 */
/* 
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.01 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.php.net/license/3_01.txt                                  |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
   | Author: Basant Kukreja <basantk@php.net>                             |
   +----------------------------------------------------------------------+
 */

#ifndef _ZEND_ATOMIC_H
#define _ZEND_ATOMIC_H

#ifdef __sun
#include <atomic.h>

#define zend_atomic_inc_32 atomic_inc_32
#define zend_atomic_inc_uint atomic_inc_uint 
#define zend_atomic_dec_32_nv atomic_dec_32_nv
#define zend_atomic_cas_ptr atomic_cas_ptr 
inline static void zend_atomic_init()
{
}

#else 
#ifndef PHP_WIN32
typedef unsigned int uint32_t;
typedef unsigned int uint_t;

#ifdef HAVE_GCC_ATOMICS
inline static void zend_atomic_init()
{
}

inline static void zend_atomic_inc_32(volatile uint32_t *target)
{
	(void) __sync_fetch_and_add(target, 1);
}

inline static void zend_atomic_inc_uint(volatile uint_t *target)
{
	(void) __sync_fetch_and_add(target, 1);
}

inline static uint32_t zend_atomic_dec_32_nv(volatile uint32_t *target)
{
	return __sync_sub_and_fetch(target, 1);
}

inline static void *zend_atomic_cas_ptr(volatile void *target, void *cmp,
								   void *newval)
{
	return __sync_val_compare_and_swap ((void**) target, cmp, newval);
}
#else

/* GCC doesn't have atomic operation so use pthread mutex instead */
#include <pthread.h>
/* XXX: This declaration should be moved to "C" file" */
static pthread_mutex_t atomic_mutex = PTHREAD_MUTEX_INITIALIZER;

inline static void zend_atomic_init()
{
	pthread_mutex_init(&atomic_mutex, NULL);
}

inline static void zend_atomic_inc_32(volatile uint32_t *target)
{
	pthread_mutex_lock(&atomic_mutex);
	++*target;
	pthread_mutex_unlock(&atomic_mutex);
}

inline static void zend_atomic_inc_uint(volatile uint_t *target)
{
	pthread_mutex_lock(&atomic_mutex);
	++*target;
	pthread_mutex_unlock(&atomic_mutex);
}

inline static uint32_t zend_atomic_dec_32_nv(volatile uint32_t *target)
{
	uint32_t newval = 0;
	pthread_mutex_lock(&atomic_mutex);
	newval = --*target;
	pthread_mutex_unlock(&atomic_mutex);
	return newval;
}

inline static void *zend_atomic_cas_ptr(volatile void *target, void *cmp,
								   void *newval)
{
	void* oldval = NULL;
	pthread_mutex_lock(&atomic_mutex);
	oldval = *(void**)target;
	if (oldval == cmp) {
		*(void**) target = newval;
	}
	pthread_mutex_unlock(&atomic_mutex);
	return oldval;
}
#endif

#else
#include "win32/php_stdint.h"
typedef LONG uint_t;

inline static void zend_atomic_init()
{
}

inline static void zend_atomic_inc_32(volatile uint32_t *target)
{
	InterlockedIncrement(target);
}

inline static void zend_atomic_inc_uint(volatile uint_t *target)
{
	InterlockedIncrement(target);
}

inline static uint32_t zend_atomic_dec_32_nv(volatile uint32_t *target)
{
	uint32_t val = InterlockedDecrement(target);
	return val - 1;
}

inline static void *zend_atomic_cas_ptr(volatile void *target, void *cmp,
								   void *newval)
{
	return InterlockedCompareExchangePointer((void**) target, newval, cmp);
}
#endif

#endif

#endif

